"""
SPARQL Burger - A Python SPARQL query builder for programmatically generating SPARQL graph patterns and queries.
Version 0.1
Official webpage: http://pmitzias.com/SPARQLBurger
Documentation: http://pmitzias.com/SPARQLBurger/docs.html
Created by Panos Mitzias (http://www.pmitzias.com)
Powered by Catalink Ltd (http://catalink.eu)
"""


class Prefix:
    def __init__(self, prefix, namespace):
        """
        The Prefix class constructor.
        :param prefix: <str> The prefix (e.g. "ex").
        :param namespace: <str> The namespace (e.g. "http://www.example.com#").
        """
        self.prefix = prefix
        self.namespace = namespace

    def get_text(self):
        """
        Generates the text for the given prefix (e.g. "PREFIX ex: <http://www.example.com#>")
        :return: <str> The prefix definition text. Returns empty string if an exception was raised.
        """
        try:
            return "PREFIX %s: <%s>\n" % (self.prefix, self.namespace)
        except Exception as e:
            print("Error 1 @ Prefix.get_text()")
            return ""


class Triple:
    def __init__(self, subject, predicate, object):
        """
        The Triple class constructor.
        :param subject: <str> The subject string (e.g. "?person")
        :param predicate: <str> The predicate string (e.g. "ex:hasName")
        :param object: <str> The object string (e.g. "\'John\'@en")
        """
        self.subject = str(subject)
        self.predicate = str(predicate)
        self.object = str(object)

    def get_text(self):
        """
        Generates the text for the given triple.
        :return: <str> The triple definition text. Returns empty string if an exception was raised.
        """
        try:
            return "%s %s %s . \n" % (self.subject, self.predicate, self.object)
        except Exception as e:
            print("Error 1 @ Triple.get_text()")
            return ""


class Filter:
    def __init__(self, expression):
        """
        The Filter class constructor.
        :param expression: <str> The expression to get in the filter (e.g. "?age > 30")
        """
        self.expression = expression

    def get_text(self):
        """
        Generates the text for the given filter.
        :return: <str> The filter definition text. Returns empty string if an exception was raised.
        """
        try:
            return "FILTER (%s)" % (self.expression,)
        except Exception as e:
            print("Error 1 @ Filter.get_text()")
            return ""

class Having:
    def __init__(self, expression):
        """
        The Having class constructor.
        :param expression: <str> The expression to get in the having filter (e.g. "?age > 30")
        """
        self.expression = expression

    def get_text(self):
        """
        Generates the text for the given having filter.
        :return: <str> The filter definition text. Returns empty string if an exception was raised.
        """
        try:
            return "HAVING (%s)" % (self.expression,)
        except Exception as e:
            print("Error 1 @ Filter.get_text()")
            return ""

class Binding:
    def __init__(self, value, variable):
        """
        The Binding class constructor.
        :param value: <str> A string value to get in the BIND first part (e.g. "John")
         OR <obj> Another object (e.g. IfClause) to be nested.
        :param variable: <str> The variable to be bound to this value (e.g. "?name")
        """
        self.value = value
        self.variable = variable

    def get_text(self):
        """
        Generates the text for the given binding (e.g. "BIND('John' AS ?name)" or "BIND(IF(BOUND(...)) AS ?name") )
        :return: <str> The binding definition text. Returns empty string if an exception was raised.
        """
        try:
            if type(self.value) is str:
                value_text = self.value
            else:
                value_text = self.value.get_text()

            return "BIND (%s AS %s)" % (value_text, self.variable)

        except Exception as e:
            print("Error 1 @ Binding.get_text()")
            return ""


class Bound:
    def __init__(self, variable):
        """
        The Bound class constructor.
        :param variable: <str> The variable to be checked if it is bound (e.g. "?name")
         OR <obj> Another object to be nested.
        """
        self.variable = variable

    def get_text(self):
        """
        Generates the text for the given BOUND clause (e.g. "BOUND (?name)" )
        :return: <str> The bound definition text. Returns empty string if an exception was raised.
        """
        try:
            if type(self.variable) is str:
                variable_text = self.variable
            else:
                variable_text = self.variable.get_text()

            return "BOUND (%s)" % (variable_text, )

        except Exception as e:
            print("Error 1 @ Bound.get_text()")
            return ""


class IfClause:
    def __init__(self, condition, true_value, false_value):
        """
        The IfClause class constructor.
        :param condition: <str> The condition for the IF clause OR <obj> Another object to be nested.
        :param true_value: <str> The value for when IF condition is True OR <obj> Another object to be nested.
        :param false_value: <str> The value for when IF condition is False OR <obj> Another object to be nested.
        """
        self.condition = condition
        self.true_value = true_value
        self.false_value = false_value

    def get_text(self):
        """
        Generates the text for the given BOUND clause (e.g. "IF(?age > 18, 'adult', 'minor')" )
        :return: <str> The if clause definition text. Returns empty string if an exception was raised.
        """
        try:
            # Check for nested condition (e.g. a nested if condition)
            if type(self.condition) is str:
                condition_text = self.condition
            else:
                condition_text = self.condition.get_text()

            # Check for nested value (e.g. a nested if value)
            if type(self.true_value) is str:
                true_value_text = self.true_value
            else:
                true_value_text = self.true_value.get_text()

            # Check for nested value (e.g. a nested if value)
            if type(self.false_value) is str:
                false_value_text = self.false_value
            else:
                false_value_text = self.false_value.get_text()

            return "IF (%s, %s, %s)" % (condition_text, true_value_text, false_value_text)

        except Exception as e:
            print("Error 1 @ IfClause.get_text()")
            return ""


class GroupBy:
    def __init__(self, variables):
        """
        The GroupBy class constructor.
        :param variables: <list> A list of variables as strings that will be used for the grouping
        """
        self.variables = variables

    def get_text(self):
        """
        Generates the text for the given GROUP BY expression (e.g. "GROUP BY ?person ?age")
        :return: <str> The GROUP BY definition text. Returns empty string if an exception was raised.
        """
        try:
            return "GROUP BY %s" % (" ".join(self.variables), )

        except Exception as e:
            print("Error 1 @ GroupBy.get_text()")
            return ""


class Values:
    def __init__(self, values, name):
        """
        The Values class constructor.
        :param values: <list> A list of variables as strings that should be
                                gathered under the same variable.
        :param name: <str> The name of the resulting variable.
        """
        self.values = values
        self.name = name

    def get_text(self):
        """
        Generate the text for the given VALUES expression (e.g.
        "VALUES ?person {<"https://www.wikidata.org/entity/42">}
        :return: <str> The VALUES defenition text. Returns empty string if an
                        exception was raised.
        """
        try:
            enclosed_values = [in_brackets(value) for value in self.values]
            return "VALUES %s {%s}" % (self.name, ' '.join(enclosed_values))
        except Exception as e:
            print("Error 1 @ Values.get_text()")
            return ""


def in_brackets(uri: str) -> str:
    """Encloses a given URI in brackets (i.e. "<" and ">").
    If the uri already has brackets, nothing happens.
    :returns: <str> A URI string enclosed in brackets.
    """
    if uri.startswith("<"):
        return uri
    elif uri.startswith("http"):
        return "<%s>" % uri
    else:
        return uri



"""
SPARQL Burger - A Python SPARQL query builder for programmatically generating SPARQL graph patterns and queries.
Version 0.1
Official webpage: http://pmitzias.com/SPARQLBurger
Documentation: http://pmitzias.com/SPARQLBurger/docs.html
Created by Panos Mitzias (http://www.pmitzias.com)
Powered by Catalink Ltd (http://catalink.eu)
"""


from SPARQLBurger.SPARQLSyntaxTerms import *


class SPARQLGraphPattern:
    def __init__(self, optional=False, union=False):
        """
        The SPARQLGraphPattern class constructor.
        :param optional: <bool> Indicates if graph pattern should be marked as OPTIONAL.
        :param union: <bool> Indicates if graph pattern should have a UNION clause that associates it with the previous.
        graph pattern
        """
        self.is_optional = optional
        self.is_union = union
        self.graph = []
        self.filters = []
        self.bindings = []
        self.values = []

    def add_triples(self, triples):
        """
        Adds a list of triples to the graph pattern.
        :param triples: <list> A list of SPARQLSyntaxTerms.Triple objects.
        :return: <bool> True if addition succeeded, False if given argument was not a list of Triple objects.
        """
        if type(triples) is list and all(isinstance(element, Triple) for element in triples):
            self.graph.extend(triples)
            return True
        else:
            return False

    def add_nested_graph_pattern(self, graph_pattern):
        """
        Adds another graph pattern as nested to the main graph pattern.
        :param graph_pattern: <obj> The SPARQLGraphPattern object to be nested.
        :return: <bool> True if addition succeeded, False if given argument was not a SPARQLGraphPattern object.
        """
        if type(graph_pattern) is SPARQLGraphPattern:
            self.graph.append(graph_pattern)
            return True
        else:
            return False

    def add_nested_select_query(self, select_query):
        """
        Adds a select query as nested to the main graph pattern.
        :param select_query: <obj> The SPARQLSelectQuery object to be nested.
        :return: <bool> True if addition succeeded, False if given argument was not a SPARQLGraphPattern object.
        """
        if type(select_query) is SPARQLSelectQuery:
            self.graph.append(select_query)
            return True
        else:
            return False

    def add_filter(self, filter):
        """
        Adds a FILTER expression to the graph pattern.
        :param filter: <obj> The Filter to be added.
        :return: <bool> True if addition succeeded, False if given argument was not a Filter object.
        """
        if type(filter) is Filter:
            self.filters.append(filter)
            return True
        else:
            return False

    def add_having(self, filter):
        """
        Adds a HAVING expression to the graph pattern.
        :param filter: <obj> The HAVING expression to be added.
        :return: <bool> True if addition succeeded, False if given argument was not a Having object.
        """
        if type(filter) is Having:
            self.filters.append(filter)
            return True
        else:
            return False

    def add_binding(self, binding):
        """
        Adds a BIND expression to the graph pattern.
        :param binding: <obj> The Binding object to be added.
        :return: <bool> True if addition succeeded, False if given argument was not a Binding object.
        """
        if type(binding) is Binding:
            self.bindings.append(binding)
            return True
        else:
            return False

    def add_value(self, value):
        """
        Adds a VALUES expression to the graph pattern.
        :param value: <Value> A Value object to be added.
        :return: <bool> True if addition succeeded, False otherwise.
        """
        if isinstance(value, Values):
            self.values.append(value)
            return True
        else:
            return False

    def get_text(self, indentation_depth=0):
        """
        Generates the text for the SPARQL graph pattern.
        :param indentation_depth: <int> A value that facilitates the appropriate addition of indents to the text. Defaults at 0.
        :return: <str> The SPARQL graph pattern text. Returns empty string if an exception was raised.
        """
        try:
            # Calculate indentations
            outer_indentation = indentation_depth * "   "
            inner_indentation = (indentation_depth + 1) * "   "

            # Initialize string
            if self.is_optional:
                query_text = "%sOPTIONAL {\n" % (outer_indentation, )
            elif self.is_union:
                query_text = "%sUNION\n%s{\n" % (outer_indentation, outer_indentation)
            else:
                query_text = "%s{\n" % (outer_indentation, )

            for value in self.values:
                query_text += "%s%s\n" % (inner_indentation, value.get_text())

            # Add triples
            for entry in self.graph:
                # If entry is a Triple object
                if type(entry) is Triple:
                    query_text += "%s%s" % (inner_indentation, entry.get_text())

                # If entry is a nested SPARQLGraphPattern object
                elif type(entry) is SPARQLGraphPattern:

                    # Get text for nested graph pattern
                    nested_graph_text = entry.get_text(indentation_depth=indentation_depth + 1)

                    # Append nested text to graph text
                    if nested_graph_text:
                        query_text += nested_graph_text
                    else:
                        return False

                # If entry is a nested SPARQLSelectQuery object
                elif type(entry) is SPARQLSelectQuery:

                    # Get the text for the nested select query
                    nested_select_text = entry.get_text(indentation_depth=indentation_depth + 2)

                    # Append nested text to graph text
                    if nested_select_text:
                        query_text += "%s{%s%s}\n" % (inner_indentation, nested_select_text, inner_indentation)
                    else:
                        return False

            # Add binding texts
            for binding in self.bindings:
                query_text += "%s%s\n" % (inner_indentation, binding.get_text())

            # Add filter texts
            for filter in self.filters:
                query_text += "%s%s\n" % (inner_indentation, filter.get_text())

            # Finalize graph text
            query_text += "%s}\n" % (outer_indentation, )

            return query_text

        except Exception as e:
            print("Error 1 @ SPARQLGraphPattern.get_text()", e)
            return ""


class SPARQLQuery:
    def __init__(self, include_popular_prefixes=False):
        """
        The SPARQLQuery class constructor.
        :param include_popular_prefixes: <bool> If True, a list of popular namespaces will be added automatically
        """
        self.prefixes = []
        self.where = None

        if include_popular_prefixes:
            self.add_popular_prefixes()

    def add_prefix(self, prefix):
        """
        Adds a PREFIX expression to the query.
        :param prefix: <obj> The Prefix object to be added.
        :return: <bool> True if addition succeeded, False if given argument was not a Prefix object.
        """
        if type(prefix) is Prefix:
            self.prefixes.append(prefix)
            return True
        else:
            return False

    def add_popular_prefixes(self):
        popular_prefixes = {
            "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "xml": "http://www.w3.org/2001/XMLSchema#",
            "owl": "http://www.w3.org/2002/07/owl#",
            "prov": "http://www.w3.org/ns/prov#",
            "foaf": "http://xmlns.com/foaf/0.1/"
        }

        for prefix in popular_prefixes:
            self.add_prefix(
                prefix=Prefix(
                    prefix=prefix,
                    namespace=popular_prefixes[prefix]
                )
            )

    def set_where_pattern(self, graph_pattern):
        """
        Sets the SPARQLGraphPattern object to be used at the WHERE part
        :param graph_pattern: <obj> The SPARQLGraphPattern object to be used.
        :return: <bool> True if setting succeeded, False if given argument was not a SPARQLGraphPattern object.
        """
        if type(graph_pattern) is SPARQLGraphPattern:
            self.where = graph_pattern
            return True
        else:
            return False


class SPARQLSelectQuery(SPARQLQuery):
    def __init__(self, distinct=False, limit=False, include_popular_prefixes=False):
        """
        The SPARQLSelectQuery class constructor.
        :param distinct: <bool> Indicates if the select should be SELECT DISTINCT.
        :param limit: <int> A limit to be used for the select query results.
        """
        SPARQLQuery.__init__(self, include_popular_prefixes)

        self.distinct = distinct
        self.limit = limit
        self.variables = []
        self.group_by = []

    def add_variables(self, variables):
        """
        Adds a list of variables to be selected by the select query
        :param variables: <list> A list of variables as strings.
        :return: <bool> True if addition succeeded, False if given argument was not a list of strings.
        """
        if type(variables) is list and all(isinstance(element, str) for element in variables):
            self.variables.extend(variables)
            return True
        else:
            return False

    def add_group_by(self, group):
        """
        Adds a GROUP BY expression to the query
        :param group: <obj> The GroupBy object to be added
        :return: <bool> True if addition succeeded, False if given argument was not a GroupBy object.
        """
        if type(group) is GroupBy:
            self.group_by.append(group)
            return True
        else:
            return False

    def get_text(self, indentation_depth=0):
        """
        Generates the text for the SPARQL select query.
        :param indentation_depth: <int> A value that facilitates the appropriate addition of indents to the text. Defaults at 0.
        :return: <str> The SPARQL Select query text. Returns empty string if an exception was raised.
        """
        try:
            # Calculate indentation
            outer_indentation = indentation_depth * "   "

            # Initialize text string
            query_text = ""

            # Add prefixes
            for prefix in self.prefixes:
                query_text += prefix.get_text()

            # Add SELECT token
            if self.distinct:
                distinct_token = "DISTINCT "
            else:
                distinct_token = ""
            query_text += "\n%sSELECT %s" % (outer_indentation, distinct_token)

            # If some variables have been defined, add them
            if self.variables:
                query_text += " ".join(self.variables)

            # If no variable has been defined, use *
            else:
                query_text += " *"

            # Add WHERE token
            query_text += "\n%sWHERE " % (outer_indentation, )

            # Add WHERE pattern graph
            if self.where is not None:
                query_text += self.where.get_text(indentation_depth=indentation_depth)[:-1]

            # Add group by expressions
            for group in self.group_by:
                query_text += "\n%s%s" % (outer_indentation, group.get_text())

            # Add limit if required
            if self.limit:
                query_text += "\nLIMIT %s" % (str(self.limit))

            return query_text

        except Exception as e:
            print("Error 1 @ SPARQLSelectQuery.get_text()", e)
            return ""


class SPARQLUpdateQuery(SPARQLQuery):
    def __init__(self, include_popular_prefixes=False):
        """
        The SPARQLUpdateQuery class constructor.
        """
        SPARQLQuery.__init__(self, include_popular_prefixes)
        self.delete = None
        self.insert = None

    def set_delete_pattern(self, graph_pattern):
        """
        Sets the SPARQLGraphPattern object to be used at the DELETE part
        :param graph_pattern: <obj> The SPARQLGraphPattern object to be used.
        :return: <bool> True if setting succeeded, False if given argument was not a SPARQLGraphPattern object.
        """
        if type(graph_pattern) is SPARQLGraphPattern:
            self.delete = graph_pattern
            return True
        else:
            return False

    def set_insert_pattern(self, graph_pattern):
        """
        Sets the SPARQLGraphPattern object to be used at the INSERT part.
        :param graph_pattern: <obj> The SPARQLGraphPattern object to be used.
        :return: <bool> True if setting succeeded, False if given argument was not a SPARQLGraphPattern object.
        """
        if type(graph_pattern) is SPARQLGraphPattern:
            self.insert = graph_pattern
            return True
        else:
            return False

    def get_text(self, indentation_depth=0):
        """
        Generates the text for the SPARQL update query.
        :param indentation_depth: <int> A value that facilitates the appropriate addition of indents to the text. Defaults at 0.
        :return: <str> The SPARQL Update query text. Returns empty string if an exception was raised.
        """

        try:
            # Calculate indentation
            outer_indentation = indentation_depth * "   "

            # Initialize text string
            query_text = ""

            # Add prefixes
            for prefix in self.prefixes:
                query_text += prefix.get_text()

            # If a delete graph pattern has been defined
            if self.delete is not None:

                # Add DELETE token
                query_text += "\n%sDELETE " % (outer_indentation,)

                # Add DELETE pattern graph
                query_text += self.delete.get_text(indentation_depth=indentation_depth)[:-1]

            # If an insert graph pattern has been defined
            if self.insert is not None:
                # Add INSERT token
                query_text += "\n%sINSERT " % (outer_indentation,)

                # Add INSERT pattern graph
                query_text += self.insert.get_text(indentation_depth=indentation_depth)[:-1]

            # If a where graph pattern has been defined
            if self.where is not None:
                # Add WHERE token
                query_text += "\n%sWHERE " % (outer_indentation,)

                # Add WHERE pattern graph
                query_text += self.where.get_text(indentation_depth=indentation_depth)[:-1]

            return query_text

        except Exception as e:
            print("Error 1 @ SPARQLUpdateQuery.get_text()", e)
            return ""



import re

from SPARQLBurger.SPARQLQueryBuilder import SPARQLGraphPattern, SPARQLSelectQuery, SPARQLUpdateQuery
from SPARQLBurger.SPARQLSyntaxTerms import Triple, Binding, IfClause, Filter, Bound, \
    Prefix, GroupBy, Values


class TestSparqlQueryBuilder:
    def test_simple_pattern(self):
        pattern = SPARQLGraphPattern()

        # Add a couple of triples to the pattern
        pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="rdf:type", object="ex:Person"),
                Triple(subject="?person", predicate="ex:hasName", object="?name")
            ]
        )

        assert generate_assert_string(pattern) == \
               "{\n ?person rdf:type ex:Person . \n ?person ex:hasName ?name . \n}\n"

    def test_optional_pattern(self):
        main_pattern = SPARQLGraphPattern()
        main_pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="rdf:type", object="ex:Person"),
                Triple(subject="?person", predicate="ex:hasName", object="?name")
            ]
        )

        optional_pattern = SPARQLGraphPattern(optional=True)
        optional_pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="ex:hasAge", object="?age")
            ]
        )

        main_pattern.add_nested_graph_pattern(optional_pattern)

        assert generate_assert_string(main_pattern) == \
               "{\n ?person rdf:type ex:Person . \n ?person ex:hasName ?name . \n" \
               " OPTIONAL {\n ?person ex:hasAge ?age . \n }\n}\n"

    def test_binding_clause(self):
        main_pattern = SPARQLGraphPattern()

        first_pattern = SPARQLGraphPattern()
        first_pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="rdf:type", object="ex:Person"),
                Triple(subject="?person", predicate="ex:hasName", object="?name")
            ]
        )

        second_pattern = SPARQLGraphPattern(union=True)
        second_pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="rdf:type", object="ex:User"),
                Triple(subject="?person", predicate="ex:hasNickname", object="?name")
            ]
        )

        main_pattern.add_nested_graph_pattern(graph_pattern=first_pattern)
        main_pattern.add_nested_graph_pattern(graph_pattern=second_pattern)

        assert generate_assert_string(main_pattern) == \
            "{\n {\n ?person rdf:type ex:Person . \n ?person ex:hasName ?name . \n }\n UNION\n" \
            " {\n ?person rdf:type ex:User . \n ?person ex:hasNickname ?name . \n }\n}\n"

    def test_values(self):
        pattern = SPARQLGraphPattern()

        uris = ["https://www.wikidata.org/entity/42",
                "https://www.wikidata.org/entity/108"]
        pattern.add_value(value=Values(values=uris, name="?friend"))

        pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="rdf:type", object="ex:Person"),
                Triple(subject="?person", predicate="foaf:knows", object="?friend")
            ]
        )

        assert generate_assert_string(pattern) == \
               "{\n VALUES ?friend {<https://www.wikidata.org/entity/42> " \
               "<https://www.wikidata.org/entity/108>}\n" \
               " ?person rdf:type ex:Person . \n ?person foaf:knows ?friend . \n}\n"

    def test_filter_bind_and_if_clauses(self):
        pattern = SPARQLGraphPattern()
        pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="rdf:type", object="ex:Person"),
                Triple(subject="?person", predicate="ex:hasAge", object="?age")
            ]
        )

        # Add a filter for variable ?age
        pattern.add_filter(
            filter=Filter(
                expression="?age < 65"
            )
        )

        # Add a binding for variable ?years_alive
        pattern.add_binding(
            binding=Binding(
                value="?age",
                variable="?years_alive"
            )
        )

        # Add a binding for variable ?status, that should be "minor" or "adult" based on the ?age value
        pattern.add_binding(
            binding=Binding(
                value=IfClause(
                    condition="?age >= 18",
                    true_value="'adult'",
                    false_value="'minor'"
                ),
                variable="?status"
            )
        )

        assert generate_assert_string(pattern) == \
            "{\n ?person rdf:type ex:Person . \n ?person ex:hasAge ?age . \n" \
            " BIND (?age AS ?years_alive)\n" \
            " BIND (IF (?age >= 18, 'adult', 'minor') AS ?status)\n" \
            " FILTER (?age < 65)\n}\n"

    def test_bind_with_recursive_nesting(self):
        pattern = SPARQLGraphPattern()
        pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="rdf:type", object="ex:Person"),
            ]
        )

        # Create an optional graph pattern and add a triple
        optional_pattern = SPARQLGraphPattern(optional=True)
        optional_pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="ex:hasAddress", object="?address")
            ]
        )

        # Add a binding with nested a IF clause and a BOUND condition
        pattern.add_binding(
            binding=Binding(
                value=IfClause(
                    condition=Bound(
                        variable="?address"
                    ),
                    true_value="?address",
                    false_value="'Unknown'"
                ),
                variable="?address"
            )
        )

        assert generate_assert_string(pattern) == \
            "{\n ?person rdf:type ex:Person . \n BIND (IF (BOUND (?address), ?address, 'Unknown') AS ?address)\n}\n"

    def test_sparql_select_query(self):
        select_query = SPARQLSelectQuery(distinct=True, limit=100)

        # Add a prefix
        select_query.add_prefix(
            prefix=Prefix(prefix="ex", namespace="http://www.example.com#")
        )

        # Add the variables we want to select
        select_query.add_variables(variables=["?person", "?age"])

        # Create a graph pattern to use for the WHERE part and add some triples
        where_pattern = SPARQLGraphPattern()
        where_pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="rdf:type", object="ex:Person"),
                Triple(subject="?person", predicate="ex:hasAge", object="?age"),
                Triple(subject="?person", predicate="ex:address", object="?address"),
            ]
        )

        # Set this graph pattern to the WHERE part
        select_query.set_where_pattern(graph_pattern=where_pattern)

        # Group the results by age
        select_query.add_group_by(
            group=GroupBy(
                variables=["?age"]
            )
        )

        assert generate_assert_string(select_query) == \
            "PREFIX ex: <http://www.example.com#>\n\nSELECT DISTINCT ?person ?age\nWHERE {\n" \
            " ?person rdf:type ex:Person . \n ?person ex:hasAge ?age . \n ?person ex:address ?address . \n}\n" \
            "GROUP BY ?age\n" \
            "LIMIT 100"

    def test_sparql_update_query(self):
        update_query = SPARQLUpdateQuery()

        # Add a prefix
        update_query.add_prefix(
            prefix=Prefix(prefix="ex", namespace="http://www.example.com#")
        )

        # Create a graph pattern for the DELETE part and add a triple
        delete_pattern = SPARQLGraphPattern()
        delete_pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="ex:hasAge", object="?age")
            ]
        )

        # Create a graph pattern for the INSERT part and add a triple
        insert_pattern = SPARQLGraphPattern()
        insert_pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="ex:hasAge", object="32")
            ]
        )

        # Create a graph pattern for the WHERE part and add some triples
        where_pattern = SPARQLGraphPattern()
        where_pattern.add_triples(
            triples=[
                Triple(subject="?person", predicate="rdf:type", object="ex:Person"),
                Triple(subject="?person", predicate="ex:hasAge", object="?age")
            ]
        )

        # Now let's append these graph patterns to our query
        update_query.set_delete_pattern(graph_pattern=delete_pattern)
        update_query.set_insert_pattern(graph_pattern=insert_pattern)
        update_query.set_where_pattern(graph_pattern=where_pattern)

        assert generate_assert_string(update_query) == \
            "PREFIX ex: <http://www.example.com#>\n\nDELETE {\n ?person ex:hasAge ?age . \n}\n" \
            "INSERT {\n ?person ex:hasAge 32 . \n}\n" \
            "WHERE {\n ?person rdf:type ex:Person . \n ?person ex:hasAge ?age . \n}"


def generate_assert_string(sparql_pattern) -> str:
    """ Returns the string representation of the given pattern.
        Shrinks any multi-space to a single space.
    """
    return re.sub(r' {2,}', ' ', sparql_pattern.get_text())
